#include "MESH_DrawerCoreVisitor.h"
#include <GL/gl.h>
#include <iostream>

using namespace std;

namespace X3DTK {
namespace MESH {

DrawerCoreVisitor::DrawerCoreVisitor()
{
  // Enter and leave functions.
  define(Recorder<Transform>::getEnterFunction(&DrawerCoreVisitor::enterTransform));
  define(Recorder<Mesh>::getEnterFunction(&DrawerCoreVisitor::enterMesh));
  define(Recorder<Transform>::getLeaveFunction(&DrawerCoreVisitor::leaveTransform));
}

void DrawerCoreVisitor::enterTransform(Transform *T)
{
  glMatrixMode(GL_MODELVIEW);
  Singleton<DrawerStateVariables>::getInstance()->pushMatrix(T->getTransform());
  glPushMatrix();
  glMultMatrixf(T->getTransform().toFloat16());
}

void DrawerCoreVisitor::leaveTransform(Transform *)
{
  glMatrixMode(GL_MODELVIEW);
  Singleton<DrawerStateVariables>::getInstance()->popMatrix();
  glPopMatrix();
}

void DrawerCoreVisitor::enterMesh(Mesh *M)
{
  // StateVariables assignation.
  DrawerStateVariables *stateVariables = Singleton<DrawerStateVariables>::getInstance();

  stateVariables->storeMesh(M);
  int meshId = stateVariables->getCurrentMeshId();

  if (stateVariables->getSelectionMode())
    glPushName(meshId);

  bool colorPerVertex = M->data().getColorPerVertex();
  bool normalPerVertex = M->data().getNormalPerVertex();
  
  switch (stateVariables->getPrimitiveType())
  {
    case 1 : // MeshViewer::VERTICES
    {
      glColor3f(1.0, 0.9, 0.8);
      glPointSize(6.0);
      const Mesh::MFVertex &vertices = M->getVertices();

      if (stateVariables->getSelectionMode())
      {
        glPushName(1);
        for (Mesh::MFVertex::const_iterator v = vertices.begin(); v != vertices.end(); ++v)
        {
          glPushName((*v)->getIndex());
          SFPoint3f point = (*v)->data().getPoint();
          glRasterPos3fv(point);
          glPopName();
        }
        glPopName();
      }
      else
      {
        glBegin(GL_POINTS);
        for (Mesh::MFVertex::const_iterator v = vertices.begin(); v != vertices.end(); ++v)
        {
          SFPoint3f point = (*v)->data().getPoint();
          glVertex3fv(point);
        }
        glEnd();
      }
      break;
    }
    
    case 2 : // MeshViewer::EDGES
    {
      glLineWidth(2.0);
      const Mesh::MFEdge &edges = M->getEdges();

      if (stateVariables->getSelectionMode())
        glPushName(2);

      int index = 0;
      for (Mesh::MFEdge::const_iterator e = edges.begin(); e != edges.end(); ++e)
      {
        if (stateVariables->getSelectionMode())
          glPushName(index++);

        if ((*e)->isBoundary())
          glColor3f(0.3, 0.3, 0.9);
        else
          glColor3f(0.7, 0.7, 0.7);

        // Normal
        SFFace *firstFace = (*((*e)->getLeftFaces().begin()));
         
        if (!firstFace)
          firstFace = (*((*e)->getRightFaces().begin()));
        SFVec3f normal;
        if (normalPerVertex)
          normal = (*e)->getFromVertex()->data().getNormalOfFace(firstFace);
        else
          normal = firstFace->data().getNormal();
         
        glNormal3fv(normal);
	   
        glBegin(GL_LINES);
        SFPoint3f point = (*e)->getFromVertex()->data().getPoint();
        glVertex3fv(point);
        point = (*e)->getToVertex()->data().getPoint();
        glVertex3fv(point);
        glEnd();
	    
        if (stateVariables->getSelectionMode())
          glPopName();
      }
	
      if (stateVariables->getSelectionMode())
        glPopName();
	
      break;
    }

    case 3 : // MeshViewer::FACES
    {
      const Mesh::MFFace& faces = M->getFaces();
	
      if (stateVariables->getSelectionMode())
        glPushName(3);

      if (stateVariables->getOneColorPerMesh())
      {
        SFColorRGBA color = SFColorRGBA(0.1*((3+meshId*3)%10), 0.1*((5+meshId*4)%10), 0.1*((7+meshId*5)%10), 1.0);
        glColor4f(color.r, color.g, color.b, color.a);
      }
	
      for (Mesh::MFFace::const_iterator f = faces.begin(); f != faces.end(); ++f)
      {
        if (!normalPerVertex)
        {
          SFVec3f normal = (*f)->data().getNormal();
          glNormal3fv(normal);
        }

        if ((!colorPerVertex) && (!stateVariables->getOneColorPerMesh()))
        {
          SFColorRGBA color = (*f)->data().getColor();
          glColor4f(color.r, color.g, color.b, color.a);
        }
	    
	if (stateVariables->getSelectionMode())
          glPushName((*f)->getIndex());

        glBegin(GL_POLYGON);
        const SFFace::MFEdge &edges = (*f)->getEdges();
        for (SFFace::MFEdge::const_iterator e = edges.begin(); e != edges.end(); ++e)
        {
          if (normalPerVertex)
          {
            SFVec3f normal = (*e)->getFromVertex()->data().getNormalOfFace(*f);
            glNormal3fv(normal);
          }
          if ((colorPerVertex) && (!stateVariables->getOneColorPerMesh()))
          {
            SFColorRGBA color = (*e)->getFromVertex()->data().getColorOfFace(*f);
            glColor4f(color.r, color.g, color.b, color.a);
          }
          
          SFPoint3f point = (*e)->getFromVertex()->data().getPoint();
          glVertex3fv(point);
        }
        glEnd();

        if (stateVariables->getSelectionMode())
          glPopName();
      }

      if (stateVariables->getSelectionMode())
        glPopName();

      break;
    }
  }

  if (stateVariables->getSelectionMode())
    glPopName();
}

}
}
